#!/usr/bin/env python3

import random
import sys
import re


class Graph:

    def __init__(self, A):
        """initialise un graphe à partir de ses listes d'adjacences `A`"""
        self.Adj = tuple(A)

    def __len__(self):
        return len(self.Adj)

    def __getitem__(self, idx):
        return self.Adj[idx]

    def randomize(self):
        for i in range(len(self.Adj)):
            random.shuffle(self.Adj[i])

    @classmethod
    def random_graph(cls, n, min_d, max_d, min_w, max_w, symmetric=False):
        G = [{} for _ in range(n)]
        for s in range(n):
            d = random.randint(min_d, max_d)
            for _ in range(d):
                v = random.randrange(n)
                p = random.uniform(min_w, max_w)
                G[s][v] = p
                if symmetric:
                    G[v][s] = p
        return cls([[(G[s][v], v) for v in G[s]] for s in range(n)])

    @property
    def N(self):
        return len(self.Adj)

    @classmethod
    def empty(cls, n):
        """initialise un graphe vide à partir de son nombre de sommets"""
        return cls([[] for _ in range(n)])

    @classmethod
    def pred2graph(cls, Pred):
        M = cls.empty(len(Pred))
        for v1, v2 in enumerate(Pred):
            if v2 is not None:
                M[v1].append(v2)
                M[v2].append(v1)
        return M

    @classmethod
    def from_file(cls, filename):
        """initialise un graphe à partir d'un fichier"""
        A = []
        f = open(filename, encoding="utf-8-sig")
        s = -1
        for L in f:
            if re.search(r"^\s*#", L):
                # comment
                continue

            s += 1
            A.append([])

            if L.strip() == "-":
                # no neighboors
                continue

            for e in L.split(';'):
                try:
                    p, v = e.split(',')
                    A[s].append((float(p), int(v)))
                except ValueError:
                    raise ValueError(f"invalid line: '{L}' (line {int(v)+1})")
        f.close()
        return cls(A)

    def to_file(self, filename, header=None, precision=3):
        if filename != "-":
            f = open(filename, mode="w", encoding="utf-8")
        else:
            f = sys.stdout

        if header:
            f.write("# "+header+"\n")
        for L in self:
            if not L:
                f.write("-\n")
                continue

            if L:
                sep = ""
                for p, v in L:
                    f.write(sep)
                    f.write("{p:.{precision:}g},{v:}"
                            .format(p=p, v=v, precision=precision))
                    sep = " ; "
            else:
                f.write("-")
            f.write("\n")
        f.close()

    def __str__(self):
        return super().__str__()


class Grid(Graph):

    def __init__(self, A, width, height):
        assert len(A) == width * height
        self.Adj = tuple(A)
        self.width = width
        self.height = height

    @classmethod
    def empty(cls, width, height):
        return cls([[] for _ in range(width*height)], width, height)

    def idx(self, x, y):
        return x + self.width * y

    def xy(self, idx):
        return idx % self.width, idx // self.width

    @classmethod
    def from_file(cls, filename, weight=(0, 1, 2, 3)):
        f = open(filename, encoding="utf-8-sig")
        M = []
        width = None
        for L in f:
            L = L.strip()
            if L:
                M.append(L.strip().replace(' ', '.'))
                if width is not None and width != len(L):
                    raise ValueError(f"file '{filename}' contains lines of different lengths")
                width = len(L)
        return cls.from_matrix(M, weight=weight)

    @classmethod
    def random_grid(cls, width, height, cells=".<>v^#", weight=(0, 1, 2, 3), savefile=None):
        M = []
        for c in cells:
            if c not in ".<>v^#":
                raise ValueError(f"'{c}' is not a valid cell")
        for i in range(height):
            M.append("".join([random.choice(cells) for _ in range(width)]))

        if savefile is not None:
            if savefile == "-":
                f = sys.stdout
            else:
                f = open(savefile, mode="w", encoding="ASCII")
            for L in M:
                f.write(L)
                f.write("\n")
            f.close()
        return cls.from_matrix(M, weight=weight)

    @classmethod
    def from_matrix(cls, M, weight=(0, 1, 2, 3)):
        """initialise un graphe à partir d'une grille"""

        height = len(M)
        width = len(M[0])
        G = cls.empty(width, height)
        for x in range(width):
            for y in range(height):
                s = G.idx(x, y)
                if x > 0:
                    v = G.idx(x-1, y)
                    if M[y][x] == '#':
                        pass    # no edge
                    elif M[y][x] == '<':
                        G[s].append((weight[0], v))
                    elif M[y][x] == '>':
                        G[s].append((weight[3], v))
                    elif M[y][x] == '.':
                        G[s].append((weight[1], v))
                    else:
                        G[s].append((weight[2], v))
                if x < width-1:
                    v = G.idx(x+1, y)
                    if M[y][x] == '#':
                        pass    # no edge
                    elif M[y][x] == '>':
                        G[s].append((weight[0], v))
                    elif M[y][x] == '<':
                        G[s].append((weight[3], v))
                    elif M[y][x] == '.':
                        G[s].append((weight[1], v))
                    else:
                        G[s].append((weight[2], v))
                if y > 0:
                    v = G.idx(x, y-1)
                    if M[y][x] == '#':
                        pass    # no edge
                    elif M[y][x] == '^':
                        G[s].append((weight[0], v))
                    elif M[y][x] == 'v':
                        G[s].append((weight[3], v))
                    elif M[y][x] == '.':
                        G[s].append((weight[1], v))
                    else:
                        G[s].append((weight[2], v))
                if y < height-1:
                    v = G.idx(x, y+1)
                    if M[y][x] == '#':
                        pass    # no edge
                    elif M[y][x] == 'v':
                        G[s].append((weight[0], v))
                    elif M[y][x] == '^':
                        G[s].append((weight[3], v))
                    elif M[y][x] == '.':
                        G[s].append((weight[1], v))
                    else:
                        G[s].append((weight[2], v))
        grid = Grid(G.Adj, width, height)
        grid.grid = M
        return grid
